# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from google.api_core.client_options import ClientOptions
from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud.aiplatform_v1 import PredictionServiceClient

from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID, GoogleBaseHook

if TYPE_CHECKING:
    from google.api_core.retry import Retry
    from google.cloud.aiplatform_v1.types import PredictResponse


class PredictionServiceHook(GoogleBaseHook):
    """Hook for Google Cloud Vertex AI Prediction API."""

    def get_prediction_service_client(self, region: str | None = None) -> PredictionServiceClient:
        """
        Return PredictionServiceClient object.

        :param region: The ID of the Google Cloud region that the service belongs to. Default is None.

        :return: `google.cloud.aiplatform_v1.services.prediction_service.client.PredictionServiceClient` instance.
        """
        if region and region != "global":
            client_options = ClientOptions(api_endpoint=f"{region}-aiplatform.googleapis.com:443")
        else:
            client_options = ClientOptions()

        return PredictionServiceClient(
            credentials=self.get_credentials(), client_info=CLIENT_INFO, client_options=client_options
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def predict(
        self,
        endpoint_id: str,
        instances: list[str],
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        parameters: dict[str, str] | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> PredictResponse:
        """
        Perform an online prediction and returns the prediction result in the response.

        :param endpoint_id: Name of the endpoint_id requested to serve the prediction.
        :param instances: Required. The instances that are the input to the prediction call. A DeployedModel
            may have an upper limit on the number of instances it supports per request, and when it is
            exceeded the prediction call errors in case of AutoML Models, or, in case of customer created
            Models, the behaviour is as documented by that Model.
        :param parameters: Additional domain-specific parameters, any string must be up to 25000 characters long.
        :param project_id: ID of the Google Cloud project where model is located if None then
            default project_id is used.
        :param location: The location of the project.
        :param retry: A retry object used to retry requests. If `None` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            `retry` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_prediction_service_client(location)
        endpoint = f"projects/{project_id}/locations/{location}/endpoints/{endpoint_id}"
        return client.predict(
            request={"endpoint": endpoint, "instances": instances, "parameters": parameters},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for Google Kubernetes Engine.
"""

from __future__ import annotations

import os
from datetime import datetime

from kubernetes.client import models as k8s

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.kubernetes_engine import (
    GKECreateClusterOperator,
    GKECreateCustomResourceOperator,
    GKEDeleteClusterOperator,
    GKEStartKueueInsideClusterOperator,
    GKEStartKueueJobOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
DAG_ID = "kubernetes_engine_kueue"
GCP_PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")

GCP_LOCATION = "europe-west3"
CLUSTER_NAME = f"gke-kueue-{ENV_ID}".replace("_", "-")
CLUSTER = {"name": CLUSTER_NAME, "initial_node_count": 1, "autopilot": {"enabled": True}}

flavor_conf = """
apiVersion: kueue.x-k8s.io/v1beta1
kind: ResourceFlavor
metadata:
  name: default-flavor
"""
cluster_conf = """
apiVersion: kueue.x-k8s.io/v1beta1
kind: ClusterQueue
metadata:
  name: cluster-queue
spec:
  namespaceSelector: {}
  queueingStrategy: BestEffortFIFO
  resourceGroups:
  - coveredResources: ["cpu", "memory", "nvidia.com/gpu", "ephemeral-storage"]
    flavors:
    - name: "default-flavor"
      resources:
      - name: "cpu"
        nominalQuota: 10
      - name: "memory"
        nominalQuota: 10Gi
      - name: "nvidia.com/gpu"
        nominalQuota: 10
      - name: "ephemeral-storage"
        nominalQuota: 10Gi
"""
QUEUE_NAME = "local-queue"
local_conf = f"""
apiVersion: kueue.x-k8s.io/v1beta1
kind: LocalQueue
metadata:
  namespace: default # LocalQueue under team-a namespace
  name: {QUEUE_NAME}
spec:
  clusterQueue: cluster-queue # Point to the ClusterQueue
"""

with DAG(
    DAG_ID,
    schedule="@once",  # Override to match your needs
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "kubernetes-engine", "kueue"],
) as dag:
    create_cluster = GKECreateClusterOperator(
        task_id="create_cluster",
        project_id=GCP_PROJECT_ID,
        location=GCP_LOCATION,
        body=CLUSTER,
    )

    # [START howto_operator_gke_install_kueue]
    add_kueue_cluster = GKEStartKueueInsideClusterOperator(
        task_id="add_kueue_cluster",
        project_id=GCP_PROJECT_ID,
        location=GCP_LOCATION,
        cluster_name=CLUSTER_NAME,
        kueue_version="v0.6.2",
    )
    # [END howto_operator_gke_install_kueue]

    create_resource_flavor = GKECreateCustomResourceOperator(
        task_id="create_resource_flavor",
        project_id=GCP_PROJECT_ID,
        location=GCP_LOCATION,
        cluster_name=CLUSTER_NAME,
        yaml_conf=flavor_conf,
        custom_resource_definition=True,
        namespaced=False,
    )

    create_cluster_queue = GKECreateCustomResourceOperator(
        task_id="create_cluster_queue",
        project_id=GCP_PROJECT_ID,
        location=GCP_LOCATION,
        cluster_name=CLUSTER_NAME,
        yaml_conf=cluster_conf,
        custom_resource_definition=True,
        namespaced=False,
    )
    create_local_queue = GKECreateCustomResourceOperator(
        task_id="create_local_queue",
        project_id=GCP_PROJECT_ID,
        location=GCP_LOCATION,
        cluster_name=CLUSTER_NAME,
        yaml_conf=local_conf,
        custom_resource_definition=True,
    )

    # [START howto_operator_kueue_start_job]
    kueue_job_task = GKEStartKueueJobOperator(
        task_id="kueue_job_task",
        project_id=GCP_PROJECT_ID,
        location=GCP_LOCATION,
        cluster_name=CLUSTER_NAME,
        queue_name=QUEUE_NAME,
        namespace="default",
        parallelism=3,
        image="perl:5.34.0",
        cmds=["perl", "-Mbignum=bpi", "-wle", "print bpi(2000)"],
        name="test-pi",
        suspend=True,
        container_resources=k8s.V1ResourceRequirements(
            requests={
                "cpu": 1,
                "memory": "200Mi",
            },
        ),
    )

    # [END howto_operator_kueue_start_job]

    delete_cluster = GKEDeleteClusterOperator(
        task_id="delete_cluster",
        cluster_name=CLUSTER_NAME,
        project_id=GCP_PROJECT_ID,
        location=GCP_LOCATION,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        # TEST SETUP
        create_cluster
        >> add_kueue_cluster
        >> create_resource_flavor
        >> create_cluster_queue
        >> create_local_queue
        # TEST BODY
        >> kueue_job_task
        # TEST TEARDOWN
        >> delete_cluster
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.videointelligence_v1 import Feature

from airflow.providers.google.cloud.hooks.video_intelligence import CloudVideoIntelligenceHook
from airflow.providers.google.common.consts import CLIENT_INFO

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_default_project_id

INPUT_URI = "gs://bucket-name/input-file"
OUTPUT_URI = "gs://bucket-name/output-file"

FEATURES = [Feature.LABEL_DETECTION]

ANNOTATE_VIDEO_RESPONSE = {"test": "test"}


class TestCloudVideoIntelligenceHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.cloud.hooks.video_intelligence.CloudVideoIntelligenceHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = CloudVideoIntelligenceHook(gcp_conn_id="test")

    @mock.patch(
        "airflow.providers.google.cloud.hooks.video_intelligence.CloudVideoIntelligenceHook.get_credentials"
    )
    @mock.patch("airflow.providers.google.cloud.hooks.video_intelligence.VideoIntelligenceServiceClient")
    def test_video_intelligence_service_client_creation(self, mock_client, mock_get_creds):
        result = self.hook.get_conn()
        mock_client.assert_called_once_with(credentials=mock_get_creds.return_value, client_info=CLIENT_INFO)
        assert mock_client.return_value == result
        assert self.hook._conn == result

    @mock.patch("airflow.providers.google.cloud.hooks.video_intelligence.CloudVideoIntelligenceHook.get_conn")
    def test_annotate_video(self, get_conn):
        # Given
        annotate_video_method = get_conn.return_value.annotate_video
        get_conn.return_value.annotate_video.return_value = ANNOTATE_VIDEO_RESPONSE

        # When
        result = self.hook.annotate_video(input_uri=INPUT_URI, features=FEATURES)

        # Then
        assert result is ANNOTATE_VIDEO_RESPONSE
        annotate_video_method.assert_called_once_with(
            request={
                "input_uri": INPUT_URI,
                "input_content": None,
                "features": FEATURES,
                "video_context": None,
                "output_uri": None,
                "location_id": None,
            },
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch("airflow.providers.google.cloud.hooks.video_intelligence.CloudVideoIntelligenceHook.get_conn")
    def test_annotate_video_with_output_uri(self, get_conn):
        # Given
        annotate_video_method = get_conn.return_value.annotate_video
        get_conn.return_value.annotate_video.return_value = ANNOTATE_VIDEO_RESPONSE

        # When
        result = self.hook.annotate_video(input_uri=INPUT_URI, output_uri=OUTPUT_URI, features=FEATURES)

        # Then
        assert result is ANNOTATE_VIDEO_RESPONSE
        annotate_video_method.assert_called_once_with(
            request={
                "input_uri": INPUT_URI,
                "output_uri": OUTPUT_URI,
                "input_content": None,
                "features": FEATURES,
                "video_context": None,
                "location_id": None,
            },
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Dataflow links."""

from __future__ import annotations

from airflow.providers.google.cloud.links.base import BaseGoogleLink

DATAFORM_BASE_LINK = "/bigquery/dataform"
DATAFORM_WORKFLOW_INVOCATION_LINK = (
    DATAFORM_BASE_LINK
    + "/locations/{region}/repositories/{repository_id}/workflows/"
    + "{workflow_invocation_id}?project={project_id}"
)
DATAFORM_REPOSITORY_LINK = (
    DATAFORM_BASE_LINK
    + "/locations/{region}/repositories/{repository_id}/"
    + "details/workspaces?project={project_id}"
)
DATAFORM_WORKSPACE_LINK = (
    DATAFORM_BASE_LINK
    + "/locations/{region}/repositories/{repository_id}/"
    + "workspaces/{workspace_id}/"
    + "files/?project={project_id}"
)


class DataformWorkflowInvocationLink(BaseGoogleLink):
    """Helper class for constructing Dataflow Job Link."""

    name = "Dataform Workflow Invocation"
    key = "dataform_workflow_invocation_config"
    format_str = DATAFORM_WORKFLOW_INVOCATION_LINK


class DataformRepositoryLink(BaseGoogleLink):
    """Helper class for constructing Dataflow repository link."""

    name = "Dataform Repository"
    key = "dataform_repository"
    format_str = DATAFORM_REPOSITORY_LINK


class DataformWorkspaceLink(BaseGoogleLink):
    """Helper class for constructing Dataform workspace link."""

    name = "Dataform Workspace"
    key = "dataform_workspace"
    format_str = DATAFORM_WORKSPACE_LINK



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Dataform sensor."""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.dataform import DataformHook
from airflow.providers.google.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import BaseSensorOperator
else:
    from airflow.sensors.base import BaseSensorOperator  # type: ignore[no-redef]

if TYPE_CHECKING:
    from airflow.utils.context import Context


class DataformWorkflowInvocationStateSensor(BaseSensorOperator):
    """
    Checks for the status of a Workflow Invocation in Google Cloud Dataform.

    :param project_id: Required, the Google Cloud project ID in which to start a job.
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param region: Required, The location of the Dataform workflow invocation (for example europe-west1).
    :param repository_id: Required. The ID of the Dataform repository that the task belongs to.
    :param workflow_invocation_id: Required, ID of the workflow invocation to be checked.
    :param expected_statuses: The expected state of the operation.
        See:
        https://cloud.google.com/python/docs/reference/dataform/latest/google.cloud.dataform_v1beta1.types.WorkflowInvocation.State
    :param failure_statuses: State that will terminate the sensor with an exception
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = ("workflow_invocation_id",)

    def __init__(
        self,
        *,
        project_id: str,
        region: str,
        repository_id: str,
        workflow_invocation_id: str,
        expected_statuses: set[int] | int,
        failure_statuses: Iterable[int] | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.repository_id = repository_id
        self.workflow_invocation_id = workflow_invocation_id
        self.expected_statuses = (
            {expected_statuses} if isinstance(expected_statuses, int) else expected_statuses
        )
        self.failure_statuses = failure_statuses
        self.project_id = project_id
        self.region = region
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.hook: DataformHook | None = None

    def poke(self, context: Context) -> bool:
        self.hook = DataformHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        workflow_invocation = self.hook.get_workflow_invocation(
            project_id=self.project_id,
            region=self.region,
            repository_id=self.repository_id,
            workflow_invocation_id=self.workflow_invocation_id,
        )
        workflow_status = workflow_invocation.state
        if workflow_status is not None:
            if self.failure_statuses and workflow_status in self.failure_statuses:
                message = (
                    f"Workflow Invocation with id '{self.workflow_invocation_id}' "
                    f"state is: {workflow_status}. Terminating sensor..."
                )
                raise AirflowException(message)

        return workflow_status in self.expected_statuses


class DataformWorkflowInvocationActionStateSensor(BaseSensorOperator):
    """
    Checks for the status of a Workflow Invocation Action in Google Cloud Dataform.

    :param project_id: Required, the Google Cloud project ID in which to start a job.
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param region: Required, The location of the Dataform workflow invocation (for example europe-west1).
    :param repository_id: Required. The ID of the Dataform repository that the task belongs to.
    :param workflow_invocation_id: Required, ID of the workflow invocation to be checked.
    :param target_name: Required. The name of the target to be checked in the workflow.
    :param expected_statuses: The expected state of the action.
        See:
        https://cloud.google.com/python/docs/reference/dataform/latest/google.cloud.dataform_v1beta1.types.WorkflowInvocationAction.State
    :param failure_statuses: State that will terminate the sensor with an exception
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = ("workflow_invocation_id",)

    def __init__(
        self,
        *,
        project_id: str,
        region: str,
        repository_id: str,
        workflow_invocation_id: str,
        target_name: str,
        expected_statuses: Iterable[int],
        failure_statuses: Iterable[int],
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.repository_id = repository_id
        self.workflow_invocation_id = workflow_invocation_id
        self.project_id = project_id
        self.region = region
        self.target_name = target_name
        self.expected_statuses = expected_statuses
        self.failure_statuses = failure_statuses
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.hook: DataformHook | None = None

    def poke(self, context: Context) -> bool:
        self.hook = DataformHook(gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain)

        workflow_invocation_actions = self.hook.query_workflow_invocation_actions(
            project_id=self.project_id,
            region=self.region,
            repository_id=self.repository_id,
            workflow_invocation_id=self.workflow_invocation_id,
        )

        for workflow_invocation_action in workflow_invocation_actions:
            if workflow_invocation_action.target.name == self.target_name:
                state = workflow_invocation_action.state
                if state in self.failure_statuses:
                    raise AirflowException(
                        f"Workflow Invocation Action target {self.target_name} state is: {state}."
                    )
                return state in self.expected_statuses

        raise AirflowException(f"Workflow Invocation Action target {self.target_name} not found.")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from google.cloud import aiplatform
from google.cloud.aiplatform.compat.types import execution_v1 as gca_execution

from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID, GoogleBaseHook


class ExperimentHook(GoogleBaseHook):
    """Use the Vertex AI SDK for Python to manage your experiments."""

    @GoogleBaseHook.fallback_to_default_project_id
    def create_experiment(
        self,
        experiment_name: str,
        location: str,
        experiment_description: str = "",
        project_id: str = PROVIDE_PROJECT_ID,
        experiment_tensorboard: str | None = None,
    ):
        """
        Create an experiment and, optionally, associate a Vertex AI TensorBoard instance using the Vertex AI SDK for Python.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param experiment_name: Required. The name of the evaluation experiment.
        :param experiment_description: Optional. Description of the evaluation experiment.
        :param experiment_tensorboard: Optional. The Vertex TensorBoard instance to use as a backing
            TensorBoard for the provided experiment. If no TensorBoard is provided, a default Tensorboard
            instance is created and used by this experiment.
        """
        aiplatform.init(
            experiment=experiment_name,
            experiment_description=experiment_description,
            experiment_tensorboard=experiment_tensorboard if experiment_tensorboard else False,
            project=project_id,
            location=location,
        )
        self.log.info("Created experiment with name: %s", experiment_name)

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_experiment(
        self,
        experiment_name: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        delete_backing_tensorboard_runs: bool = False,
    ) -> None:
        """
        Delete an experiment.

        Deleting an experiment deletes that experiment and all experiment runs associated with the experiment.
        The Vertex AI TensorBoard experiment associated with the experiment is not deleted.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param experiment_name: Required. The name of the evaluation experiment.
        :param delete_backing_tensorboard_runs: Optional. If True will also delete the Vertex AI TensorBoard
            runs associated with the experiment runs under this experiment that we used to store time series
            metrics.
        """
        experiment = aiplatform.Experiment(
            experiment_name=experiment_name, project=project_id, location=location
        )

        experiment.delete(delete_backing_tensorboard_runs=delete_backing_tensorboard_runs)


class ExperimentRunHook(GoogleBaseHook):
    """Use the Vertex AI SDK for Python to create and manage your experiment runs."""

    @GoogleBaseHook.fallback_to_default_project_id
    def create_experiment_run(
        self,
        experiment_run_name: str,
        experiment_name: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        experiment_run_tensorboard: str | None = None,
        run_after_creation: bool = False,
    ) -> None:
        """
        Create experiment run for the experiment.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param experiment_name: Required. The name of the evaluation experiment.
        :param experiment_run_name: Required. The specific run name or ID for this experiment.
        :param experiment_run_tensorboard: Optional. A backing TensorBoard resource to enable and store time
            series metrics logged to this experiment run.
        :param run_after_creation: Optional. Responsible for state after creation of experiment run.
            If true experiment run will be created with state RUNNING.
        """
        experiment_run_state = (
            gca_execution.Execution.State.NEW
            if not run_after_creation
            else gca_execution.Execution.State.RUNNING
        )
        experiment_run = aiplatform.ExperimentRun.create(
            run_name=experiment_run_name,
            experiment=experiment_name,
            project=project_id,
            location=location,
            state=experiment_run_state,
            tensorboard=experiment_run_tensorboard,
        )
        self.log.info(
            "Created experiment run with name: %s and status: %s",
            experiment_run.name,
            experiment_run.state,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def list_experiment_runs(
        self,
        experiment_name: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
    ) -> list[aiplatform.ExperimentRun]:
        """
        List experiment run for the experiment.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param experiment_name: Required. The name of the evaluation experiment.
        """
        experiment_runs = aiplatform.ExperimentRun.list(
            experiment=experiment_name,
            project=project_id,
            location=location,
        )
        return experiment_runs

    @GoogleBaseHook.fallback_to_default_project_id
    def update_experiment_run_state(
        self,
        experiment_run_name: str,
        experiment_name: str,
        location: str,
        new_state: gca_execution.Execution.State,
        project_id: str = PROVIDE_PROJECT_ID,
    ) -> None:
        """
        Update state of the experiment run.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param experiment_name: Required. The name of the evaluation experiment.
        :param experiment_run_name: Required. The specific run name or ID for this experiment.
        :param new_state: Required. New state of the experiment run.
        """
        experiment_run = aiplatform.ExperimentRun(
            run_name=experiment_run_name,
            experiment=experiment_name,
            project=project_id,
            location=location,
        )
        self.log.info("State of the %s before update is: %s", experiment_run.name, experiment_run.state)

        experiment_run.update_state(new_state)

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_experiment_run(
        self,
        experiment_run_name: str,
        experiment_name: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        delete_backing_tensorboard_run: bool = False,
    ) -> None:
        """
        Delete experiment run from the experiment.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param experiment_name: Required. The name of the evaluation experiment.
        :param experiment_run_name: Required. The specific run name or ID for this experiment.
        :param delete_backing_tensorboard_run: Whether to delete the backing Vertex AI TensorBoard run
            that stores time series metrics for this run.
        """
        self.log.info("Next experiment run will be deleted: %s", experiment_run_name)
        experiment_run = aiplatform.ExperimentRun(
            run_name=experiment_run_name, experiment=experiment_name, project=project_id, location=location
        )
        experiment_run.delete(delete_backing_tensorboard_run=delete_backing_tensorboard_run)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Hooks for Cloud Memorystore service.

.. spelling:word-list::

    DataProtectionMode
    FieldMask
    pb
    memcache
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from google.api_core import path_template
from google.api_core.exceptions import NotFound
from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud.memcache_v1beta2 import CloudMemcacheClient
from google.cloud.memcache_v1beta2.types import cloud_memcache
from google.cloud.redis_v1 import (
    CloudRedisClient,
    FailoverInstanceRequest,
    InputConfig,
    Instance,
    OutputConfig,
)

from airflow import version
from airflow.exceptions import AirflowException
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID, GoogleBaseHook

if TYPE_CHECKING:
    from google.api_core.retry import Retry
    from google.protobuf.field_mask_pb2 import FieldMask


class CloudMemorystoreHook(GoogleBaseHook):
    """
    Hook for Google Cloud Memorystore APIs.

    All the methods in the hook where project_id is used must be called with
    keyword arguments rather than positional.

    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account.
    """

    def __init__(
        self,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            gcp_conn_id=gcp_conn_id,
            impersonation_chain=impersonation_chain,
            **kwargs,
        )
        self._client: CloudRedisClient | None = None

    def get_conn(self) -> CloudRedisClient:
        """Retrieve client library object that allow access to Cloud Memorystore service."""
        if not self._client:
            self._client = CloudRedisClient(credentials=self.get_credentials())
        return self._client

    @staticmethod
    def _append_label(instance: Instance, key: str, val: str) -> Instance:
        """
        Append labels to provided Instance type.

        Labels must fit the regex ``[a-z]([-a-z0-9]*[a-z0-9])?`` (current
         airflow version string follows semantic versioning spec: x.y.z).

        :param instance: The proto to append resource_label airflow
            version to
        :param key: The key label
        :param val:
        :return: The cluster proto updated with new label
        """
        val = val.replace(".", "-").replace("+", "-")
        instance.labels.update({key: val})
        return instance

    @GoogleBaseHook.fallback_to_default_project_id
    def create_instance(
        self,
        location: str,
        instance_id: str,
        instance: dict | Instance,
        project_id: str = PROVIDE_PROJECT_ID,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Create a Redis instance based on the specified tier and memory size.

        By default, the instance is accessible from the project's `default network
        <https://cloud.google.com/compute/docs/networks-and-firewalls#networks>`__.

        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance_id: Required. The logical name of the Redis instance in the customer project with the
            following restrictions:

            -  Must contain only lowercase letters, numbers, and hyphens.
            -  Must start with a letter.
            -  Must be between 1-40 characters.
            -  Must end with a number or a letter.
            -  Must be unique within the customer project / location
        :param instance: Required. A Redis [Instance] resource

            If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.cloud.redis_v1.types.Instance`
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the Google Cloud connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        if isinstance(instance, dict):
            instance = Instance(**instance)
        elif not isinstance(instance, Instance):
            raise AirflowException("instance is not instance of Instance type or python dict")

        parent = f"projects/{project_id}/locations/{location}"
        instance_name = f"projects/{project_id}/locations/{location}/instances/{instance_id}"
        try:
            self.log.info("Fetching instance: %s", instance_name)
            instance = client.get_instance(
                request={"name": instance_name}, retry=retry, timeout=timeout, metadata=metadata or ()
            )
            self.log.info("Instance exists. Skipping creation.")
            return instance
        except NotFound:
            self.log.info("Instance not exists.")

        self._append_label(instance, "airflow-version", "v" + version.version)

        result = client.create_instance(
            request={"parent": parent, "instance_id": instance_id, "instance": instance},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        result.result()
        self.log.info("Instance created.")
        return client.get_instance(
            request={"name": instance_name}, retry=retry, timeout=timeout, metadata=metadata or ()
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_instance(
        self,
        location: str,
        instance: str,
        project_id: str = PROVIDE_PROJECT_ID,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Delete a specific Redis instance.  Instance stops serving and data is deleted.

        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance: The logical name of the Redis instance in the customer project.
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the Google Cloud connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        name = f"projects/{project_id}/locations/{location}/instances/{instance}"
        self.log.info("Fetching Instance: %s", name)
        instance = client.get_instance(
            request={"name": name},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        if not instance:
            return

        self.log.info("Deleting Instance: %s", name)
        result = client.delete_instance(
            request={"name": name},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        result.result()
        self.log.info("Instance deleted: %s", name)

    @GoogleBaseHook.fallback_to_default_project_id
    def export_instance(
        self,
        location: str,
        instance: str,
        output_config: dict | OutputConfig,
        project_id: str = PROVIDE_PROJECT_ID,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Export Redis instance data into a Redis RDB format file in Cloud Storage.

        Redis will continue serving during this operation.

        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance: The logical name of the Redis instance in the customer project.
        :param output_config: Required. Specify data to be exported.

            If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.cloud.redis_v1.types.OutputConfig`
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the Google Cloud connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        name = f"projects/{project_id}/locations/{location}/instances/{instance}"
        self.log.info("Exporting Instance: %s", name)
        result = client.export_instance(
            request={"name": name, "output_config": output_config},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        result.result()
        self.log.info("Instance exported: %s", name)

    @GoogleBaseHook.fallback_to_default_project_id
    def failover_instance(
        self,
        location: str,
        instance: str,
        data_protection_mode: FailoverInstanceRequest.DataProtectionMode,
        project_id: str = PROVIDE_PROJECT_ID,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Failover of the primary node to current replica node.

        Initiates a failover of the primary node to current replica node for a specific STANDARD tier Cloud
        Memorystore for Redis instance.

        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance: The logical name of the Redis instance in the customer project.
        :param data_protection_mode: Optional. Available data protection modes that the user can choose. If
            it's unspecified, data protection mode will be LIMITED_DATA_LOSS by default.
            .DataProtectionMode
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the Google Cloud connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        name = f"projects/{project_id}/locations/{location}/instances/{instance}"
        self.log.info("Failovering Instance: %s", name)

        result = client.failover_instance(
            request={"name": name, "data_protection_mode": data_protection_mode},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        result.result()
        self.log.info("Instance failovered: %s", name)

    @GoogleBaseHook.fallback_to_default_project_id
    def get_instance(
        self,
        location: str,
        instance: str,
        project_id: str = PROVIDE_PROJECT_ID,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Get the details of a specific Redis instance.

        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance: The logical name of the Redis instance in the customer project.
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the Google Cloud connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        name = f"projects/{project_id}/locations/{location}/instances/{instance}"
        result = client.get_instance(
            request={"name": name},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        self.log.info("Fetched Instance: %s", name)
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def import_instance(
        self,
        location: str,
        instance: str,
        input_config: dict | InputConfig,
        project_id: str = PROVIDE_PROJECT_ID,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Import a Redis RDB snapshot file from Cloud Storage into a Redis instance.

        Redis may stop serving during this operation. Instance state will be IMPORTING for entire operation.
        When complete, the instance will contain only data from the imported file.

        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance: The logical name of the Redis instance in the customer project.
        :param input_config: Required. Specify data to be imported.

            If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.cloud.redis_v1.types.InputConfig`
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the Google Cloud connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        name = f"projects/{project_id}/locations/{location}/instances/{instance}"
        self.log.info("Importing Instance: %s", name)
        result = client.import_instance(
            request={"name": name, "input_config": input_config},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        result.result()
        self.log.info("Instance imported: %s", name)

    @GoogleBaseHook.fallback_to_default_project_id
    def list_instances(
        self,
        location: str,
        page_size: int,
        project_id: str = PROVIDE_PROJECT_ID,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        List Redis instances owned by a project at the specified location (region) or all locations.

        :param location: The location of the Cloud Memorystore instance (for example europe-west1)

                If it is specified as ``-`` (wildcard), then all regions available to the project are
                queried, and the results are aggregated.
        :param page_size: The maximum number of resources contained in the underlying API response. If page
            streaming is performed per- resource, this parameter does not affect the return value. If page
            streaming is performed per-page, this determines the maximum number of resources in a page.
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the Google Cloud connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        parent = f"projects/{project_id}/locations/{location}"
        result = client.list_instances(
            request={"parent": parent, "page_size": page_size},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        self.log.info("Fetched instances")
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def update_instance(
        self,
        update_mask: dict | FieldMask,
        instance: dict | Instance,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str | None = None,
        instance_id: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Update the metadata and configuration of a specific Redis instance.

        :param update_mask: Required. Mask of fields to update. At least one path must be supplied in this
            field. The elements of the repeated paths field may only include these fields from ``Instance``:

            -  ``displayName``
            -  ``labels``
            -  ``memorySizeGb``
            -  ``redisConfig``

            If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.protobuf.field_mask_pb2.FieldMask`
        :param instance: Required. Update description. Only fields specified in ``update_mask`` are updated.

            If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.cloud.redis_v1.types.Instance`
        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance_id: The logical name of the Redis instance in the customer project.
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the Google Cloud connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()

        if isinstance(instance, dict):
            instance = Instance(**instance)
        elif not isinstance(instance, Instance):
            raise AirflowException("instance is not instance of Instance type or python dict")

        if location and instance_id:
            name = f"projects/{project_id}/locations/{location}/instances/{instance_id}"
            instance.name = name

        self.log.info("Updating instances: %s", instance.name)
        result = client.update_instance(
            request={"update_mask": update_mask, "instance": instance},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        updated_instance = result.result()
        self.log.info("Instance updated: %s", instance.name)
        return updated_instance


class CloudMemorystoreMemcachedHook(GoogleBaseHook):
    """
    Hook for Google Cloud Memorystore for Memcached service APIs.

    All the methods in the hook where project_id is used must be called with
    keyword arguments rather than positional.

    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account.
    """

    def __init__(
        self,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            gcp_conn_id=gcp_conn_id,
            impersonation_chain=impersonation_chain,
            **kwargs,
        )
        self._client: CloudMemcacheClient | None = None

    def get_conn(self):
        """Retrieve client library object that allow access to Cloud Memorystore Memcached service."""
        if not self._client:
            self._client = CloudMemcacheClient(credentials=self.get_credentials())
        return self._client

    @staticmethod
    def _append_label(instance: cloud_memcache.Instance, key: str, val: str) -> cloud_memcache.Instance:
        """
        Append labels to provided Instance type.

        Labels must fit the regex ``[a-z]([-a-z0-9]*[a-z0-9])?`` (current
         airflow version string follows semantic versioning spec: x.y.z).

        :param instance: The proto to append resource_label airflow
            version to
        :param key: The key label
        :param val:
        :return: The cluster proto updated with new label
        """
        val = val.replace(".", "-").replace("+", "-")
        instance.labels.update({key: val})
        return instance

    @GoogleBaseHook.fallback_to_default_project_id
    def apply_parameters(
        self,
        node_ids: Sequence[str],
        apply_all: bool,
        project_id: str,
        location: str,
        instance_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Will update current set of Parameters to the set of specified nodes of the Memcached Instance.

        :param node_ids: Nodes to which we should apply the instance-level parameter group.
        :param apply_all: Whether to apply instance-level parameter group to all nodes. If set to true,
            will explicitly restrict users from specifying any nodes, and apply parameter group updates
            to all nodes within the instance.
        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance_id: The logical name of the Memcached instance in the customer project.
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the Google Cloud connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        metadata = metadata or ()
        name = CloudMemcacheClient.instance_path(project_id, location, instance_id)

        self.log.info("Applying update to instance: %s", instance_id)
        result = client.apply_parameters(
            name=name,
            node_ids=node_ids,
            apply_all=apply_all,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        result.result()
        self.log.info("Instance updated: %s", instance_id)

    @GoogleBaseHook.fallback_to_default_project_id
    def create_instance(
        self,
        location: str,
        instance_id: str,
        instance: dict | cloud_memcache.Instance,
        project_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Create a Memcached instance based on the specified tier and memory size.

        By default, the instance is accessible from the project's `default network
        <https://cloud.google.com/compute/docs/networks-and-firewalls#networks>`__.

        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance_id: Required. The logical name of the Memcached instance in the customer project
            with the following restrictions:

            -  Must contain only lowercase letters, numbers, and hyphens.
            -  Must start with a letter.
            -  Must be between 1-40 characters.
            -  Must end with a number or a letter.
            -  Must be unique within the customer project / location
        :param instance: Required. A Memcached [Instance] resource

            If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.cloud.memcache_v1beta2.types.cloud_memcache.Instance`
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the GCP connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        metadata = metadata or ()
        parent = path_template.expand(
            "projects/{project}/locations/{location}", project=project_id, location=location
        )
        instance_name = CloudMemcacheClient.instance_path(project_id, location, instance_id)
        try:
            instance = client.get_instance(
                name=instance_name, retry=retry, timeout=timeout, metadata=metadata
            )
            self.log.info("Instance exists. Skipping creation.")
            return instance
        except NotFound:
            self.log.info("Instance not exists.")

        if isinstance(instance, dict):
            instance = cloud_memcache.Instance(instance)
        elif not isinstance(instance, cloud_memcache.Instance):
            raise AirflowException("instance is not instance of Instance type or python dict")

        self._append_label(instance, "airflow-version", "v" + version.version)

        result = client.create_instance(
            parent=parent,
            instance_id=instance_id,
            resource=instance,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        result.result()
        self.log.info("Instance created.")
        return client.get_instance(
            name=instance_name,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_instance(
        self,
        location: str,
        instance: str,
        project_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Delete a specific Memcached instance.  Instance stops serving and data is deleted.

        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance: The logical name of the Memcached instance in the customer project.
        :param project_id:  Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the GCP connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        metadata = metadata or ()
        name = CloudMemcacheClient.instance_path(project_id, location, instance)
        self.log.info("Fetching Instance: %s", name)
        instance = client.get_instance(
            name=name,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        if not instance:
            return

        self.log.info("Deleting Instance: %s", name)
        result = client.delete_instance(
            name=name,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        result.result()
        self.log.info("Instance deleted: %s", name)

    @GoogleBaseHook.fallback_to_default_project_id
    def get_instance(
        self,
        location: str,
        instance: str,
        project_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Get the details of a specific Memcached instance.

        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance: The logical name of the Memcached instance in the customer project.
        :param project_id:  Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the GCP connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        metadata = metadata or ()
        name = CloudMemcacheClient.instance_path(project_id, location, instance)
        result = client.get_instance(name=name, retry=retry, timeout=timeout, metadata=metadata or ())
        self.log.info("Fetched Instance: %s", name)
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def list_instances(
        self,
        location: str,
        project_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        List Memcached instances owned by a project at the specified location (region) or all locations.

        :param location: The location of the Cloud Memorystore instance (for example europe-west1)

                If it is specified as ``-`` (wildcard), then all regions available to the project are
                queried, and the results are aggregated.
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the GCP connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        metadata = metadata or ()
        parent = path_template.expand(
            "projects/{project}/locations/{location}", project=project_id, location=location
        )
        result = client.list_instances(
            parent=parent,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        self.log.info("Fetched instances")
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def update_instance(
        self,
        update_mask: dict | FieldMask,
        instance: dict | cloud_memcache.Instance,
        project_id: str,
        location: str | None = None,
        instance_id: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Update the metadata and configuration of a specific Memcached instance.

        :param update_mask: Required. Mask of fields to update. At least one path must be supplied in this
            field. The elements of the repeated paths field may only include these fields from ``Instance``:

            -  ``displayName``

            If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.protobuf.field_mask_pb2.FieldMask`)
            Union[dict, google.protobuf.field_mask_pb2.FieldMask]
        :param instance: Required. Update description. Only fields specified in ``update_mask`` are updated.

            If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.cloud.memcache_v1beta2.types.cloud_memcache.Instance`
        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance_id: The logical name of the Memcached instance in the customer project.
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the Google Cloud connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        metadata = metadata or ()

        if isinstance(instance, dict):
            instance = cloud_memcache.Instance(instance)
        elif not isinstance(instance, cloud_memcache.Instance):
            raise AirflowException("instance is not instance of Instance type or python dict")

        if location and instance_id:
            name = CloudMemcacheClient.instance_path(project_id, location, instance_id)
            instance.name = name

        self.log.info("Updating instances: %s", instance.name)
        result = client.update_instance(
            update_mask=update_mask, resource=instance, retry=retry, timeout=timeout, metadata=metadata or ()
        )
        updated_instance = result.result()
        self.log.info("Instance updated: %s", instance.name)
        return updated_instance

    @GoogleBaseHook.fallback_to_default_project_id
    def update_parameters(
        self,
        update_mask: dict | FieldMask,
        parameters: dict | cloud_memcache.MemcacheParameters,
        project_id: str,
        location: str,
        instance_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Update the defined Memcached Parameters for an existing Instance.

        This method only stages the parameters, it must be followed by apply_parameters
        to apply the parameters to nodes of the Memcached Instance.

        :param update_mask: Required. Mask of fields to update.
            If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.protobuf.field_mask_pb2.FieldMask`
        :param parameters: The parameters to apply to the instance.
            If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.cloud.memcache_v1beta2.types.cloud_memcache.MemcacheParameters`
        :param location: The location of the Cloud Memorystore instance (for example europe-west1)
        :param instance_id: The logical name of the Memcached instance in the customer project.
        :param project_id: Project ID of the project that contains the instance. If set
            to None or missing, the default project_id from the Google Cloud connection is used.
        :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
            retried.
        :param timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
            ``retry`` is specified, the timeout applies to each individual attempt.
        :param metadata: Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        metadata = metadata or ()

        if isinstance(parameters, dict):
            parameters = cloud_memcache.MemcacheParameters(parameters)
        elif not isinstance(parameters, cloud_memcache.MemcacheParameters):
            raise AirflowException("instance is not instance of MemcacheParameters type or python dict")

        name = CloudMemcacheClient.instance_path(project_id, location, instance_id)
        self.log.info("Staging update to instance: %s", instance_id)
        result = client.update_parameters(
            name=name,
            update_mask=update_mask,
            parameters=parameters,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        result.result()
        self.log.info("Update staged for instance: %s", instance_id)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

# For no Pydantic environment, we need to skip the tests
pytest.importorskip("google.cloud.aiplatform_v1")

from airflow.providers.google.cloud.links.translate import (
    TRANSLATION_BASE_LINK,
    TranslationDatasetListLink,
    TranslationLegacyDatasetLink,
    TranslationLegacyModelLink,
    TranslationLegacyModelTrainLink,
)
from airflow.providers.google.cloud.operators.automl import (
    AutoMLCreateDatasetOperator,
    AutoMLListDatasetOperator,
    AutoMLTrainModelOperator,
)

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk.execution_time.comms import XComResult

GCP_LOCATION = "test-location"
GCP_PROJECT_ID = "test-project"
DATASET = "test-dataset"
MODEL = "test-model"


class TestTranslationLegacyDatasetLink:
    @pytest.mark.db_test
    def test_get_link(self, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = f"{TRANSLATION_BASE_LINK}/locations/{GCP_LOCATION}/datasets/{DATASET}/sentences?project={GCP_PROJECT_ID}"
        link = TranslationLegacyDatasetLink()
        ti = create_task_instance_of_operator(
            AutoMLCreateDatasetOperator,
            dag_id="test_legacy_dataset_link_dag",
            task_id="test_legacy_dataset_link_task",
            dataset=DATASET,
            location=GCP_LOCATION,
        )
        session.add(ti)
        session.commit()

        link.persist(
            context={"ti": ti, "task": ti.task},
            dataset_id=DATASET,
            project_id=GCP_PROJECT_ID,
            location=GCP_LOCATION,
        )

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "location": ti.task.location,
                    "dataset_id": DATASET,
                    "project_id": GCP_PROJECT_ID,
                },
            )
        actual_url = link.get_link(operator=ti.task, ti_key=ti.key)
        assert actual_url == expected_url


class TestTranslationDatasetListLink:
    @pytest.mark.db_test
    def test_get_link(self, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = f"{TRANSLATION_BASE_LINK}/datasets?project={GCP_PROJECT_ID}"
        link = TranslationDatasetListLink()
        ti = create_task_instance_of_operator(
            AutoMLListDatasetOperator,
            dag_id="test_dataset_list_link_dag",
            task_id="test_dataset_list_link_task",
            location=GCP_LOCATION,
        )
        session.add(ti)
        session.commit()
        link.persist(context={"ti": ti, "task": ti.task}, project_id=GCP_PROJECT_ID)

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "project_id": GCP_PROJECT_ID,
                },
            )
        actual_url = link.get_link(operator=ti.task, ti_key=ti.key)
        assert actual_url == expected_url


class TestTranslationLegacyModelLink:
    @pytest.mark.db_test
    def test_get_link(self, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = (
            f"{TRANSLATION_BASE_LINK}/locations/{GCP_LOCATION}/datasets/{DATASET}/"
            f"evaluate;modelId={MODEL}?project={GCP_PROJECT_ID}"
        )
        link = TranslationLegacyModelLink()
        ti = create_task_instance_of_operator(
            AutoMLTrainModelOperator,
            dag_id="test_legacy_model_link_dag",
            task_id="test_legacy_model_link_task",
            model=MODEL,
            project_id=GCP_PROJECT_ID,
            location=GCP_LOCATION,
        )
        session.add(ti)
        session.commit()
        task = mock.MagicMock()
        task.extra_links_params = {
            "dataset_id": DATASET,
            "model_id": MODEL,
            "project_id": GCP_PROJECT_ID,
            "location": GCP_LOCATION,
        }
        link.persist(
            context={"ti": ti, "task": task},
            dataset_id=DATASET,
            model_id=MODEL,
            project_id=GCP_PROJECT_ID,
        )
        if mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "location": ti.task.location,
                    "dataset_id": DATASET,
                    "model_id": MODEL,
                    "project_id": GCP_PROJECT_ID,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url


class TestTranslationLegacyModelTrainLink:
    @pytest.mark.db_test
    def test_get_link(self, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = (
            f"{TRANSLATION_BASE_LINK}/locations/{GCP_LOCATION}/datasets/{DATASET}/"
            f"train?project={GCP_PROJECT_ID}"
        )
        link = TranslationLegacyModelTrainLink()
        ti = create_task_instance_of_operator(
            AutoMLTrainModelOperator,
            dag_id="test_legacy_model_train_link_dag",
            task_id="test_legacy_model_train_link_task",
            model={"dataset_id": DATASET},
            project_id=GCP_PROJECT_ID,
            location=GCP_LOCATION,
        )
        session.add(ti)
        session.commit()
        task = mock.MagicMock()
        task.extra_links_params = {
            "dataset_id": DATASET,
            "project_id": GCP_PROJECT_ID,
            "location": GCP_LOCATION,
        }
        link.persist(
            context={"ti": ti, "task": task},
            dataset_id=DATASET,
            project_id=GCP_PROJECT_ID,
            location=GCP_LOCATION,
        )

        if mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "location": ti.task.location,
                    "dataset_id": ti.task.model["dataset_id"],
                    "project_id": GCP_PROJECT_ID,
                },
            )
        actual_url = link.get_link(operator=task, ti_key=ti.key)
        assert actual_url == expected_url



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


"""
Example Airflow DAG for Google Vertex AI service testing Dataset operations.
"""

from __future__ import annotations

import os
from datetime import datetime

from google.cloud.aiplatform import schema
from google.protobuf.json_format import ParseDict
from google.protobuf.struct_pb2 import Value

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gcs import (
    GCSCreateBucketOperator,
    GCSDeleteBucketOperator,
)
from airflow.providers.google.cloud.operators.vertex_ai.dataset import (
    CreateDatasetOperator,
    DeleteDatasetOperator,
    ExportDataOperator,
    GetDatasetOperator,
    ImportDataOperator,
    ListDatasetsOperator,
    UpdateDatasetOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
DAG_ID = "vertex_ai_dataset_operations"
REGION = "us-central1"

RESOURCE_DATA_BUCKET = "airflow-system-tests-resources"
DATA_SAMPLE_GCS_BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}".replace("_", "-")

TIME_SERIES_DATASET = {
    "display_name": f"time-series-dataset-{ENV_ID}",
    "metadata_schema_uri": schema.dataset.metadata.time_series,
    "metadata": ParseDict(
        {
            "input_config": {
                "gcs_source": {
                    "uri": [f"gs://{RESOURCE_DATA_BUCKET}/vertex-ai/datasets/forecast-dataset.csv"]
                }
            }
        },
        Value(),
    ),
}
IMAGE_DATASET = {
    "display_name": f"image-dataset-{ENV_ID}",
    "metadata_schema_uri": schema.dataset.metadata.image,
    "metadata": Value(string_value="image-dataset"),
}
TABULAR_DATASET = {
    "display_name": f"tabular-dataset-{ENV_ID}",
    "metadata_schema_uri": schema.dataset.metadata.tabular,
    "metadata": ParseDict(
        {
            "input_config": {
                "gcs_source": {"uri": [f"gs://{RESOURCE_DATA_BUCKET}/vertex-ai/datasets/tabular-dataset.csv"]}
            }
        },
        Value(),
    ),
}
VIDEO_DATASET = {
    "display_name": f"video-dataset-{ENV_ID}",
    "metadata_schema_uri": schema.dataset.metadata.video,
    "metadata": Value(string_value="video-dataset"),
}
TEST_EXPORT_CONFIG = {"gcs_destination": {"output_uri_prefix": f"gs://{DATA_SAMPLE_GCS_BUCKET_NAME}/exports"}}
TEST_IMPORT_CONFIG = [
    {
        "data_item_labels": {
            "test-labels-name": "test-labels-value",
        },
        "import_schema_uri": schema.dataset.ioformat.image.single_label_classification,
        "gcs_source": {"uris": [f"gs://{RESOURCE_DATA_BUCKET}/vertex-ai/datasets/image-dataset-flowers.csv"]},
    },
]
DATASET_TO_UPDATE = {"display_name": "test-name"}
TEST_UPDATE_MASK = {"paths": ["displayName"]}


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "vertex_ai", "dataset"],
) as dag:
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket",
        bucket_name=DATA_SAMPLE_GCS_BUCKET_NAME,
        storage_class="REGIONAL",
        location=REGION,
    )

    # [START how_to_cloud_vertex_ai_create_dataset_operator]
    create_image_dataset_job = CreateDatasetOperator(
        task_id="image_dataset",
        dataset=IMAGE_DATASET,
        region=REGION,
        project_id=PROJECT_ID,
    )
    create_tabular_dataset_job = CreateDatasetOperator(
        task_id="tabular_dataset",
        dataset=TABULAR_DATASET,
        region=REGION,
        project_id=PROJECT_ID,
    )
    create_video_dataset_job = CreateDatasetOperator(
        task_id="video_dataset",
        dataset=VIDEO_DATASET,
        region=REGION,
        project_id=PROJECT_ID,
    )
    create_time_series_dataset_job = CreateDatasetOperator(
        task_id="time_series_dataset",
        dataset=TIME_SERIES_DATASET,
        region=REGION,
        project_id=PROJECT_ID,
    )
    # [END how_to_cloud_vertex_ai_create_dataset_operator]

    # [START how_to_cloud_vertex_ai_get_dataset_operator]
    get_dataset = GetDatasetOperator(
        task_id="get_dataset",
        project_id=PROJECT_ID,
        region=REGION,
        dataset_id=create_tabular_dataset_job.output["dataset_id"],
    )
    # [END how_to_cloud_vertex_ai_get_dataset_operator]

    # [START how_to_cloud_vertex_ai_export_data_operator]
    export_data_job = ExportDataOperator(
        task_id="export_data",
        dataset_id=create_image_dataset_job.output["dataset_id"],
        region=REGION,
        project_id=PROJECT_ID,
        export_config=TEST_EXPORT_CONFIG,
    )
    # [END how_to_cloud_vertex_ai_export_data_operator]

    # [START how_to_cloud_vertex_ai_import_data_operator]
    import_data_job = ImportDataOperator(
        task_id="import_data",
        dataset_id=create_image_dataset_job.output["dataset_id"],
        region=REGION,
        project_id=PROJECT_ID,
        import_configs=TEST_IMPORT_CONFIG,
    )
    # [END how_to_cloud_vertex_ai_import_data_operator]

    # [START how_to_cloud_vertex_ai_list_dataset_operator]
    list_dataset_job = ListDatasetsOperator(
        task_id="list_dataset",
        region=REGION,
        project_id=PROJECT_ID,
    )
    # [END how_to_cloud_vertex_ai_list_dataset_operator]

    # [START how_to_cloud_vertex_ai_update_dataset_operator]
    update_dataset_job = UpdateDatasetOperator(
        task_id="update_dataset",
        project_id=PROJECT_ID,
        region=REGION,
        dataset_id=create_video_dataset_job.output["dataset_id"],
        dataset=DATASET_TO_UPDATE,
        update_mask=TEST_UPDATE_MASK,
    )
    # [END how_to_cloud_vertex_ai_update_dataset_operator]

    # [START how_to_cloud_vertex_ai_delete_dataset_operator]
    delete_dataset_job = DeleteDatasetOperator(
        task_id="delete_dataset",
        dataset_id=create_time_series_dataset_job.output["dataset_id"],
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END how_to_cloud_vertex_ai_delete_dataset_operator]

    delete_tabular_dataset_job = DeleteDatasetOperator(
        task_id="delete_tabular_dataset",
        dataset_id=create_tabular_dataset_job.output["dataset_id"],
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_image_dataset_job = DeleteDatasetOperator(
        task_id="delete_image_dataset",
        dataset_id=create_image_dataset_job.output["dataset_id"],
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_video_dataset_job = DeleteDatasetOperator(
        task_id="delete_video_dataset",
        dataset_id=create_video_dataset_job.output["dataset_id"],
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=DATA_SAMPLE_GCS_BUCKET_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        # TEST SETUP
        create_bucket
        # TEST BODY
        >> [
            create_time_series_dataset_job >> delete_dataset_job,
            create_tabular_dataset_job >> get_dataset >> delete_tabular_dataset_job,
            create_image_dataset_job >> import_data_job >> export_data_job >> delete_image_dataset_job,
            create_video_dataset_job >> update_dataset_job >> delete_video_dataset_job,
            list_dataset_job,
        ]
        # TEST TEARDOWN
        >> delete_bucket
    )

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.exceptions import AirflowException, AirflowNotFoundException
from airflow.providers.google.cloud.hooks.datafusion import PipelineStates
from airflow.providers.google.cloud.sensors.datafusion import CloudDataFusionPipelineStateSensor

LOCATION = "test-location"
INSTANCE_NAME = "airflow-test-instance"
INSTANCE_URL = "http://datafusion.instance.com"
PIPELINE_NAME = "shrubberyPipeline"
PIPELINE_ID = "test_pipeline_id"
PROJECT_ID = "test_project_id"
GCP_CONN_ID = "test_conn_id"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]
FAILURE_STATUSES = {"FAILED"}


class TestCloudDataFusionPipelineStateSensor:
    @pytest.mark.parametrize(
        "expected_status, current_status, sensor_return",
        [
            (PipelineStates.COMPLETED, PipelineStates.COMPLETED, True),
            (PipelineStates.COMPLETED, PipelineStates.RUNNING, False),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.sensors.datafusion.DataFusionHook")
    def test_poke(self, mock_hook, expected_status, current_status, sensor_return):
        mock_hook.return_value.get_instance.return_value = {"apiEndpoint": INSTANCE_URL}

        task = CloudDataFusionPipelineStateSensor(
            task_id="test_task_id",
            pipeline_name=PIPELINE_NAME,
            pipeline_id=PIPELINE_ID,
            project_id=PROJECT_ID,
            expected_statuses={expected_status},
            instance_name=INSTANCE_NAME,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_pipeline_workflow.return_value = {"status": current_status}
        result = task.poke(mock.MagicMock())

        assert sensor_return == result

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME, location=LOCATION, project_id=PROJECT_ID
        )

    @mock.patch("airflow.providers.google.cloud.sensors.datafusion.DataFusionHook")
    def test_assertion(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {"apiEndpoint": INSTANCE_URL}

        task = CloudDataFusionPipelineStateSensor(
            task_id="test_task_id",
            pipeline_name=PIPELINE_NAME,
            pipeline_id=PIPELINE_ID,
            project_id=PROJECT_ID,
            expected_statuses={PipelineStates.COMPLETED},
            failure_statuses=FAILURE_STATUSES,
            instance_name=INSTANCE_NAME,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_pipeline_workflow.return_value = {"status": "FAILED"}
        with pytest.raises(
            AirflowException,
            match=f"Pipeline with id '{PIPELINE_ID}' state is: FAILED. Terminating sensor...",
        ):
            task.poke(mock.MagicMock())

    @mock.patch("airflow.providers.google.cloud.sensors.datafusion.DataFusionHook")
    def test_not_found_exception(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {"apiEndpoint": INSTANCE_URL}
        mock_hook.return_value.get_pipeline_workflow.side_effect = AirflowNotFoundException()

        task = CloudDataFusionPipelineStateSensor(
            task_id="test_task_id",
            pipeline_name=PIPELINE_NAME,
            pipeline_id=PIPELINE_ID,
            project_id=PROJECT_ID,
            expected_statuses={PipelineStates.COMPLETED},
            failure_statuses=FAILURE_STATUSES,
            instance_name=INSTANCE_NAME,
            location=LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        with pytest.raises(
            AirflowException,
            match="Specified Pipeline ID was not found.",
        ):
            task.poke(mock.MagicMock())



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from copy import deepcopy
from typing import Any
from unittest import mock

import pytest
from google.api_core.gapic_v1.method import DEFAULT
from google.cloud import alloydb_v1

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.alloy_db import AlloyDbHook
from airflow.providers.google.common.consts import CLIENT_INFO

try:
    import importlib.util

    if not importlib.util.find_spec("airflow.sdk.bases.hook"):
        raise ImportError

    BASEHOOK_PATCH_PATH = "airflow.sdk.bases.hook.BaseHook"
except ImportError:
    BASEHOOK_PATCH_PATH = "airflow.hooks.base.BaseHook"
TEST_GCP_PROJECT = "test-project"
TEST_GCP_REGION = "global"
TEST_GCP_CONN_ID = "test_conn_id"
TEST_IMPERSONATION_CHAIN = "test_impersonation_chain"

TEST_CLUSTER_ID = "test_cluster_id"
TEST_CLUSTER: dict[str, Any] = {}
TEST_CLUSTER_NAME = f"projects/{TEST_GCP_PROJECT}/locations/{TEST_GCP_REGION}/clusters/{TEST_CLUSTER_ID}"
TEST_UPDATE_MASK = None
TEST_ALLOW_MISSING = False
TEST_ETAG = "test-etag"
TEST_FORCE = False
TEST_REQUEST_ID = "test_request_id"
TEST_VALIDATE_ONLY = False
TEST_USER_ID = "test_user"
TEST_USER: dict[str, Any] = {}
TEST_USER_NAME = f"{TEST_CLUSTER_NAME}/users/{TEST_USER_ID}"
TEST_BACKUP_ID = "test_backup_id"
TEST_BACKUP: dict[str, Any] = {}
TEST_BACKUP_NAME = f"projects/{TEST_GCP_PROJECT}/locations/{TEST_GCP_REGION}/backups/{TEST_BACKUP_ID}"
TEST_RETRY = DEFAULT
TEST_TIMEOUT = None
TEST_METADATA = ()

HOOK_PATH = "airflow.providers.google.cloud.hooks.alloy_db.{}"


class TestAlloyDbHook:
    def setup_method(self):
        with mock.patch(f"{BASEHOOK_PATCH_PATH}.get_connection"):
            self.hook = AlloyDbHook(
                gcp_conn_id=TEST_GCP_CONN_ID,
            )

    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_credentials"))
    @mock.patch(HOOK_PATH.format("alloydb_v1.AlloyDBAdminClient"))
    def test_get_alloy_db_admin_client(self, mock_client, mock_get_credentials):
        mock_credentials = mock_get_credentials.return_value
        expected_client = mock_client.return_value

        client = self.hook.get_alloy_db_admin_client()

        assert client == expected_client
        mock_get_credentials.assert_called_once()
        mock_client.assert_called_once_with(
            credentials=mock_credentials,
            client_info=CLIENT_INFO,
        )

    @pytest.mark.parametrize(
        "given_timeout, expected_timeout",
        [
            (None, None),
            (0.0, None),
            (10.0, 10),
            (10.9, 10),
        ],
    )
    @mock.patch(HOOK_PATH.format("AlloyDbHook.log"))
    def test_wait_for_operation(self, mock_log, given_timeout, expected_timeout):
        mock_operation = mock.MagicMock()
        expected_operation_result = mock_operation.result.return_value

        result = self.hook.wait_for_operation(timeout=given_timeout, operation=mock_operation)

        assert result == expected_operation_result
        mock_log.info.assert_called_once_with("Waiting for operation to complete...")
        mock_operation.result.assert_called_once_with(timeout=expected_timeout)

    @pytest.mark.parametrize(
        "given_timeout, expected_timeout",
        [
            (None, None),
            (0.0, None),
            (10.0, 10),
            (10.9, 10),
        ],
    )
    @mock.patch(HOOK_PATH.format("AlloyDbHook.log"))
    def test_wait_for_operation_exception(self, mock_log, given_timeout, expected_timeout):
        mock_operation = mock.MagicMock()
        mock_operation.result.side_effect = Exception

        with pytest.raises(AirflowException):
            self.hook.wait_for_operation(timeout=given_timeout, operation=mock_operation)

        mock_log.info.assert_called_once_with("Waiting for operation to complete...")
        mock_operation.result.assert_called_once_with(timeout=expected_timeout)
        mock_operation.exception.assert_called_once_with(timeout=expected_timeout)

    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_create_cluster(self, mock_client):
        mock_create_cluster = mock_client.return_value.create_cluster
        expected_result = mock_create_cluster.return_value
        expected_parent = f"projects/{TEST_GCP_PROJECT}/locations/{TEST_GCP_REGION}"
        mock_common_location_path = mock_client.return_value.common_location_path
        mock_common_location_path.return_value = expected_parent
        expected_request = {
            "parent": expected_parent,
            "cluster_id": TEST_CLUSTER_ID,
            "cluster": TEST_CLUSTER,
            "request_id": TEST_REQUEST_ID,
            "validate_only": TEST_VALIDATE_ONLY,
        }

        result = self.hook.create_cluster(
            cluster_id=TEST_CLUSTER_ID,
            cluster=TEST_CLUSTER,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_client.assert_called_once()
        mock_common_location_path.assert_called_once_with(TEST_GCP_PROJECT, TEST_GCP_REGION)
        mock_create_cluster.assert_called_once_with(
            request=expected_request,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_create_secondary_cluster(self, mock_client):
        mock_create_secondary_cluster = mock_client.return_value.create_secondary_cluster
        expected_result = mock_create_secondary_cluster.return_value
        expected_parent = f"projects/{TEST_GCP_PROJECT}/locations/{TEST_GCP_REGION}"
        mock_common_location_path = mock_client.return_value.common_location_path
        mock_common_location_path.return_value = expected_parent
        expected_request = {
            "parent": expected_parent,
            "cluster_id": TEST_CLUSTER_ID,
            "cluster": TEST_CLUSTER,
            "request_id": TEST_REQUEST_ID,
            "validate_only": TEST_VALIDATE_ONLY,
        }

        result = self.hook.create_secondary_cluster(
            cluster_id=TEST_CLUSTER_ID,
            cluster=TEST_CLUSTER,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_client.assert_called_once()
        mock_common_location_path.assert_called_once_with(TEST_GCP_PROJECT, TEST_GCP_REGION)
        mock_create_secondary_cluster.assert_called_once_with(
            request=expected_request,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_get_cluster(self, mock_client):
        mock_get_cluster = mock_client.return_value.get_cluster
        mock_cluster_path = mock_client.return_value.cluster_path
        mock_cluster_path.return_value = TEST_CLUSTER_NAME
        expected_result = mock_get_cluster.return_value

        result = self.hook.get_cluster(
            cluster_id=TEST_CLUSTER_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_client.assert_called_once()
        mock_cluster_path.assert_called_once_with(TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_CLUSTER_ID)
        mock_get_cluster.assert_called_once_with(
            request={"name": TEST_CLUSTER_NAME},
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @pytest.mark.parametrize(
        "given_cluster, expected_cluster",
        [
            (TEST_CLUSTER, {**deepcopy(TEST_CLUSTER), **{"name": TEST_CLUSTER_NAME}}),
            (alloydb_v1.Cluster(), {"name": TEST_CLUSTER_NAME}),
            ({}, {"name": TEST_CLUSTER_NAME}),
        ],
    )
    @mock.patch(HOOK_PATH.format("deepcopy"))
    @mock.patch(HOOK_PATH.format("alloydb_v1.Cluster.to_dict"))
    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_update_cluster(self, mock_client, mock_to_dict, mock_deepcopy, given_cluster, expected_cluster):
        mock_update_cluster = mock_client.return_value.update_cluster
        expected_result = mock_update_cluster.return_value
        mock_deepcopy.return_value = expected_cluster
        mock_to_dict.return_value = expected_cluster
        mock_cluster_path = mock_client.return_value.cluster_path
        mock_cluster_path.return_value = expected_cluster

        expected_request = {
            "update_mask": TEST_UPDATE_MASK,
            "cluster": expected_cluster,
            "request_id": TEST_REQUEST_ID,
            "validate_only": TEST_VALIDATE_ONLY,
            "allow_missing": TEST_ALLOW_MISSING,
        }

        result = self.hook.update_cluster(
            cluster_id=TEST_CLUSTER_ID,
            cluster=given_cluster,
            location=TEST_GCP_REGION,
            update_mask=TEST_UPDATE_MASK,
            project_id=TEST_GCP_PROJECT,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_cluster_path.assert_called_once_with(TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_CLUSTER_ID)
        if isinstance(given_cluster, dict):
            mock_deepcopy.assert_called_once_with(given_cluster)
            assert not mock_to_dict.called
        else:
            assert not mock_deepcopy.called
            mock_to_dict.assert_called_once_with(given_cluster)
        mock_client.assert_called_once()
        mock_update_cluster.assert_called_once_with(
            request=expected_request,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_delete_cluster(self, mock_client):
        mock_delete_cluster = mock_client.return_value.delete_cluster
        expected_result = mock_delete_cluster.return_value
        mock_cluster_path = mock_client.return_value.cluster_path
        mock_cluster_path.return_value = TEST_CLUSTER_NAME
        expected_request = {
            "name": TEST_CLUSTER_NAME,
            "request_id": TEST_REQUEST_ID,
            "etag": TEST_ETAG,
            "validate_only": TEST_VALIDATE_ONLY,
            "force": TEST_FORCE,
        }

        result = self.hook.delete_cluster(
            cluster_id=TEST_CLUSTER_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            etag=TEST_ETAG,
            validate_only=TEST_VALIDATE_ONLY,
            force=TEST_FORCE,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_client.assert_called_once()
        mock_cluster_path.assert_called_once_with(TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_CLUSTER_ID)
        mock_delete_cluster.assert_called_once_with(
            request=expected_request,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_create_user(self, mock_client):
        mock_create_user = mock_client.return_value.create_user
        expected_result = mock_create_user.return_value
        expected_parent = TEST_CLUSTER_NAME
        mock_cluster_path = mock_client.return_value.cluster_path
        mock_cluster_path.return_value = expected_parent
        expected_request = {
            "parent": expected_parent,
            "user_id": TEST_USER_ID,
            "user": TEST_USER,
            "request_id": TEST_REQUEST_ID,
            "validate_only": TEST_VALIDATE_ONLY,
        }

        result = self.hook.create_user(
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            user=TEST_USER,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_client.assert_called_once()
        mock_cluster_path.assert_called_once_with(TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_CLUSTER_ID)
        mock_create_user.assert_called_once_with(
            request=expected_request,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_get_user(self, mock_client):
        mock_get_user = mock_client.return_value.get_user
        mock_user_path = mock_client.return_value.user_path
        mock_user_path.return_value = TEST_USER_NAME
        expected_result = mock_get_user.return_value

        result = self.hook.get_user(
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_client.assert_called_once()
        mock_user_path.assert_called_once_with(
            TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_CLUSTER_ID, TEST_USER_ID
        )
        mock_get_user.assert_called_once_with(
            request={"name": TEST_USER_NAME},
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @pytest.mark.parametrize(
        "given_user, expected_user",
        [
            (TEST_USER, {**deepcopy(TEST_USER), **{"name": TEST_USER_NAME}}),
            (alloydb_v1.User(), {"name": TEST_USER_NAME}),
            ({}, {"name": TEST_USER_NAME}),
        ],
    )
    @mock.patch(HOOK_PATH.format("deepcopy"))
    @mock.patch(HOOK_PATH.format("alloydb_v1.User.to_dict"))
    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_update_user(self, mock_client, mock_to_dict, mock_deepcopy, given_user, expected_user):
        mock_update_user = mock_client.return_value.update_user
        expected_result = mock_update_user.return_value
        mock_deepcopy.return_value = expected_user
        mock_to_dict.return_value = expected_user
        mock_user_path = mock_client.return_value.user_path
        mock_user_path.return_value = expected_user

        expected_request = {
            "update_mask": TEST_UPDATE_MASK,
            "user": expected_user,
            "request_id": TEST_REQUEST_ID,
            "validate_only": TEST_VALIDATE_ONLY,
            "allow_missing": TEST_ALLOW_MISSING,
        }

        result = self.hook.update_user(
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            user=given_user,
            location=TEST_GCP_REGION,
            update_mask=TEST_UPDATE_MASK,
            project_id=TEST_GCP_PROJECT,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_user_path.assert_called_once_with(
            TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_CLUSTER_ID, TEST_USER_ID
        )
        if isinstance(given_user, dict):
            mock_deepcopy.assert_called_once_with(given_user)
            assert not mock_to_dict.called
        else:
            assert not mock_deepcopy.called
            mock_to_dict.assert_called_once_with(given_user)
        mock_client.assert_called_once()
        mock_update_user.assert_called_once_with(
            request=expected_request,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_delete_user(self, mock_client):
        mock_delete_user = mock_client.return_value.delete_user
        expected_result = mock_delete_user.return_value
        mock_user_path = mock_client.return_value.user_path
        mock_user_path.return_value = TEST_USER_NAME
        expected_request = {
            "name": TEST_USER_NAME,
            "request_id": TEST_REQUEST_ID,
            "validate_only": TEST_VALIDATE_ONLY,
        }

        result = self.hook.delete_user(
            user_id=TEST_USER_ID,
            cluster_id=TEST_CLUSTER_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_client.assert_called_once()
        mock_user_path.assert_called_once_with(
            TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_CLUSTER_ID, TEST_USER_ID
        )
        mock_delete_user.assert_called_once_with(
            request=expected_request,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_create_backup(self, mock_client):
        mock_create_backup = mock_client.return_value.create_backup
        expected_result = mock_create_backup.return_value
        expected_parent = f"projects/{TEST_GCP_PROJECT}/locations/{TEST_GCP_REGION}"
        mock_common_location_path = mock_client.return_value.common_location_path
        mock_common_location_path.return_value = expected_parent
        expected_request = {
            "parent": expected_parent,
            "backup_id": TEST_BACKUP_ID,
            "backup": TEST_BACKUP,
            "request_id": TEST_REQUEST_ID,
            "validate_only": TEST_VALIDATE_ONLY,
        }

        result = self.hook.create_backup(
            backup_id=TEST_BACKUP_ID,
            backup=TEST_BACKUP,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_client.assert_called_once()
        mock_common_location_path.assert_called_once_with(TEST_GCP_PROJECT, TEST_GCP_REGION)
        mock_create_backup.assert_called_once_with(
            request=expected_request,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_get_backup(self, mock_client):
        mock_get_backup = mock_client.return_value.get_backup
        mock_cluster_backup = mock_client.return_value.backup_path
        mock_cluster_backup.return_value = TEST_BACKUP_NAME
        expected_result = mock_get_backup.return_value

        result = self.hook.get_backup(
            backup_id=TEST_BACKUP_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_client.assert_called_once()
        mock_cluster_backup.assert_called_once_with(TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_BACKUP_ID)
        mock_get_backup.assert_called_once_with(
            request={"name": TEST_BACKUP_NAME},
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @pytest.mark.parametrize(
        "given_backup, expected_backup",
        [
            (TEST_BACKUP, {**deepcopy(TEST_BACKUP), **{"name": TEST_BACKUP_NAME}}),
            (alloydb_v1.Backup(), {"name": TEST_BACKUP_NAME}),
            ({}, {"name": TEST_BACKUP_NAME}),
        ],
    )
    @mock.patch(HOOK_PATH.format("deepcopy"))
    @mock.patch(HOOK_PATH.format("alloydb_v1.Backup.to_dict"))
    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_update_backup(self, mock_client, mock_to_dict, mock_deepcopy, given_backup, expected_backup):
        mock_update_backup = mock_client.return_value.update_backup
        expected_result = mock_update_backup.return_value
        mock_deepcopy.return_value = expected_backup
        mock_to_dict.return_value = expected_backup
        mock_backup_path = mock_client.return_value.backup_path
        mock_backup_path.return_value = expected_backup

        expected_request = {
            "update_mask": TEST_UPDATE_MASK,
            "backup": expected_backup,
            "request_id": TEST_REQUEST_ID,
            "validate_only": TEST_VALIDATE_ONLY,
            "allow_missing": TEST_ALLOW_MISSING,
        }

        result = self.hook.update_backup(
            backup_id=TEST_BACKUP_ID,
            backup=given_backup,
            location=TEST_GCP_REGION,
            update_mask=TEST_UPDATE_MASK,
            project_id=TEST_GCP_PROJECT,
            allow_missing=TEST_ALLOW_MISSING,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_backup_path.assert_called_once_with(TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_BACKUP_ID)
        if isinstance(given_backup, dict):
            mock_deepcopy.assert_called_once_with(given_backup)
            assert not mock_to_dict.called
        else:
            assert not mock_deepcopy.called
            mock_to_dict.assert_called_once_with(given_backup)
        mock_client.assert_called_once()
        mock_update_backup.assert_called_once_with(
            request=expected_request,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(HOOK_PATH.format("AlloyDbHook.get_alloy_db_admin_client"))
    def test_delete_backup(self, mock_client):
        mock_delete_backup = mock_client.return_value.delete_backup
        expected_result = mock_delete_backup.return_value
        mock_backup_path = mock_client.return_value.backup_path
        mock_backup_path.return_value = TEST_BACKUP_NAME
        expected_request = {
            "name": TEST_BACKUP_NAME,
            "request_id": TEST_REQUEST_ID,
            "validate_only": TEST_VALIDATE_ONLY,
        }

        result = self.hook.delete_backup(
            backup_id=TEST_BACKUP_ID,
            location=TEST_GCP_REGION,
            project_id=TEST_GCP_PROJECT,
            request_id=TEST_REQUEST_ID,
            validate_only=TEST_VALIDATE_ONLY,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

        assert result == expected_result
        mock_client.assert_called_once()
        mock_backup_path.assert_called_once_with(TEST_GCP_PROJECT, TEST_GCP_REGION, TEST_BACKUP_ID)
        mock_delete_backup.assert_called_once_with(
            request=expected_request,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Data Catalog links."""

from __future__ import annotations

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.providers.google.cloud.links.base import BaseGoogleLink
from airflow.providers.google.common.deprecated import deprecated

DATACATALOG_BASE_LINK = "/datacatalog"
ENTRY_GROUP_LINK = (
    DATACATALOG_BASE_LINK
    + "/groups/{entry_group_id};container={project_id};location={location_id}?project={project_id}"
)
ENTRY_LINK = (
    DATACATALOG_BASE_LINK
    + "/projects/{project_id}/locations/{location_id}/entryGroups/{entry_group_id}/entries/{entry_id}\
    ?project={project_id}"
)
TAG_TEMPLATE_LINK = (
    DATACATALOG_BASE_LINK
    + "/projects/{project_id}/locations/{location_id}/tagTemplates/{tag_template_id}?project={project_id}"
)


@deprecated(
    planned_removal_date="January 30, 2026",
    use_instead="airflow.providers.google.cloud.links.dataplex.DataplexCatalogEntryGroupLink",
    reason="The Data Catalog will be discontinued on January 30, 2026 "
    "in favor of Dataplex Universal Catalog.",
    category=AirflowProviderDeprecationWarning,
)
class DataCatalogEntryGroupLink(BaseGoogleLink):
    """Helper class for constructing Data Catalog Entry Group Link."""

    name = "Data Catalog Entry Group"
    key = "data_catalog_entry_group"
    format_str = ENTRY_GROUP_LINK


@deprecated(
    planned_removal_date="January 30, 2026",
    use_instead="airflow.providers.google.cloud.links.dataplex.DataplexCatalogEntryLink",
    reason="The Data Catalog will be discontinued on January 30, 2026 "
    "in favor of Dataplex Universal Catalog.",
    category=AirflowProviderDeprecationWarning,
)
class DataCatalogEntryLink(BaseGoogleLink):
    """Helper class for constructing Data Catalog Entry Link."""

    name = "Data Catalog Entry"
    key = "data_catalog_entry"
    format_str = ENTRY_LINK


@deprecated(
    planned_removal_date="January 30, 2026",
    use_instead="airflow.providers.google.cloud.links.dataplex.DataplexCatalogAspectTypeLink",
    reason="The Data Catalog will be discontinued on January 30, 2026 "
    "in favor of Dataplex Universal Catalog.",
    category=AirflowProviderDeprecationWarning,
)
class DataCatalogTagTemplateLink(BaseGoogleLink):
    """Helper class for constructing Data Catalog Tag Template Link."""

    name = "Data Catalog Tag Template"
    key = "data_catalog_tag_template"
    format_str = TAG_TEMPLATE_LINK



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any
from unittest import mock

import pytest

from airflow.providers.google.cloud.links.base import BaseGoogleLink
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk.execution_time.comms import XComResult

TEST_LOCATION = "test-location"
TEST_CLUSTER_ID = "test-cluster-id"
TEST_PROJECT_ID = "test-project-id"
EXPECTED_GOOGLE_LINK_KEY = "google_link_for_test"
EXPECTED_GOOGLE_LINK_NAME = "Google Link for Test"
EXPECTED_GOOGLE_LINK_FORMAT = "/services/locations/{location}/clusters/{cluster_id}?project={project_id}"
EXPECTED_GOOGLE_LINK = "https://console.cloud.google.com" + EXPECTED_GOOGLE_LINK_FORMAT.format(
    location=TEST_LOCATION, cluster_id=TEST_CLUSTER_ID, project_id=TEST_PROJECT_ID
)


class GoogleLink(BaseGoogleLink):
    key = EXPECTED_GOOGLE_LINK_KEY
    name = EXPECTED_GOOGLE_LINK_NAME
    format_str = EXPECTED_GOOGLE_LINK_FORMAT


class TestBaseGoogleLink:
    def test_class_attributes(self):
        assert GoogleLink.key == EXPECTED_GOOGLE_LINK_KEY
        assert GoogleLink.name == EXPECTED_GOOGLE_LINK_NAME
        assert GoogleLink.format_str == EXPECTED_GOOGLE_LINK_FORMAT

    def test_persist(self):
        mock_context = mock.MagicMock()

        if AIRFLOW_V_3_0_PLUS:
            GoogleLink.persist(
                context=mock_context,
                location=TEST_LOCATION,
                cluster_id=TEST_CLUSTER_ID,
                project_id=TEST_PROJECT_ID,
            )
            mock_context["ti"].xcom_push.assert_called_once_with(
                key=EXPECTED_GOOGLE_LINK_KEY,
                value={
                    "location": TEST_LOCATION,
                    "cluster_id": TEST_CLUSTER_ID,
                    "project_id": TEST_PROJECT_ID,
                },
            )
        else:
            GoogleLink.persist(
                context=mock_context,
                location=TEST_LOCATION,
                cluster_id=TEST_CLUSTER_ID,
                project_id=TEST_PROJECT_ID,
            )


class MyOperator(GoogleCloudBaseOperator):
    operator_extra_links = (GoogleLink(),)

    def __init__(self, project_id: str, location: str, cluster_id: str, **kwargs):
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.cluster_id = cluster_id

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "project_id": self.project_id,
            "cluster_id": self.cluster_id,
            "location": self.location,
        }

    def execute(self, context) -> Any:
        GoogleLink.persist(context=context)


class TestOperatorWithBaseGoogleLink:
    @pytest.mark.db_test
    def test_get_link(self, create_task_instance_of_operator, session, mock_supervisor_comms):
        expected_url = EXPECTED_GOOGLE_LINK
        link = GoogleLink()
        ti = create_task_instance_of_operator(
            MyOperator,
            dag_id="test_link_dag",
            task_id="test_link_task",
            location=TEST_LOCATION,
            cluster_id=TEST_CLUSTER_ID,
            project_id=TEST_PROJECT_ID,
        )
        session.add(ti)
        session.commit()

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key="key",
                value={
                    "cluster_id": ti.task.cluster_id,
                    "location": ti.task.location,
                    "project_id": ti.task.project_id,
                },
            )
        actual_url = link.get_link(operator=ti.task, ti_key=ti.key)
        assert actual_url == expected_url



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for Google Cloud Stackdriver service.
"""

from __future__ import annotations

import json
import os
from datetime import datetime

from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.stackdriver import (
    StackdriverDeleteAlertOperator,
    StackdriverDeleteNotificationChannelOperator,
    StackdriverDisableAlertPoliciesOperator,
    StackdriverDisableNotificationChannelsOperator,
    StackdriverEnableAlertPoliciesOperator,
    StackdriverEnableNotificationChannelsOperator,
    StackdriverListAlertPoliciesOperator,
    StackdriverListNotificationChannelsOperator,
    StackdriverUpsertAlertOperator,
    StackdriverUpsertNotificationChannelOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
DAG_ID = "stackdriver"

ALERT_1_NAME = f"alert_{ENV_ID}_1"
ALERT_2_NAME = f"alert_{ENV_ID}_2"
CHANNEL_1_NAME = f"channel_{ENV_ID}_1"
CHANNEL_2_NAME = f"channel_{ENV_ID}_2"

TEST_ALERT_POLICY_1 = {
    "combiner": "OR",
    "enabled": True,
    "display_name": ALERT_1_NAME,
    "conditions": [
        {
            "condition_threshold": {
                "filter": (
                    'metric.label.state="blocked" AND '
                    'metric.type="agent.googleapis.com/processes/count_by_state" '
                    'AND resource.type="gce_instance"'
                ),
                "comparison": "COMPARISON_GT",
                "threshold_value": 100,
                "duration": {"seconds": 900},
                "trigger": {"percent": 0},
                "aggregations": [
                    {
                        "alignment_period": {"seconds": 60},
                        "per_series_aligner": "ALIGN_MEAN",
                        "cross_series_reducer": "REDUCE_MEAN",
                        "group_by_fields": ["project", "resource.label.instance_id", "resource.label.zone"],
                    }
                ],
            },
            "display_name": f"{ALERT_1_NAME}_policy_1",
        }
    ],
}

TEST_ALERT_POLICY_2 = {
    "combiner": "OR",
    "enabled": False,
    "display_name": ALERT_2_NAME,
    "conditions": [
        {
            "condition_threshold": {
                "filter": (
                    'metric.label.state="blocked" AND '
                    'metric.type="agent.googleapis.com/processes/count_by_state" AND '
                    'resource.type="gce_instance"'
                ),
                "comparison": "COMPARISON_GT",
                "threshold_value": 100,
                "duration": {"seconds": 900},
                "trigger": {"percent": 0},
                "aggregations": [
                    {
                        "alignment_period": {"seconds": 60},
                        "per_series_aligner": "ALIGN_MEAN",
                        "cross_series_reducer": "REDUCE_MEAN",
                        "group_by_fields": ["project", "resource.label.instance_id", "resource.label.zone"],
                    }
                ],
            },
            "display_name": f"{ALERT_2_NAME}_policy_2",
        }
    ],
}

TEST_NOTIFICATION_CHANNEL_1 = {
    "display_name": CHANNEL_1_NAME,
    "enabled": True,
    "labels": {"topic": f"projects/{PROJECT_ID}/topics/notificationTopic"},
    "type": "pubsub",
}

TEST_NOTIFICATION_CHANNEL_2 = {
    "display_name": CHANNEL_2_NAME,
    "enabled": False,
    "labels": {"topic": f"projects/{PROJECT_ID}/topics/notificationTopic2"},
    "type": "pubsub",
}

with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "stackdriver"],
) as dag:
    # [START howto_operator_gcp_stackdriver_upsert_notification_channel]
    create_notification_channel = StackdriverUpsertNotificationChannelOperator(
        task_id="create-notification-channel",
        channels=json.dumps({"channels": [TEST_NOTIFICATION_CHANNEL_1, TEST_NOTIFICATION_CHANNEL_2]}),
    )
    # [END howto_operator_gcp_stackdriver_upsert_notification_channel]

    # [START howto_operator_gcp_stackdriver_enable_notification_channel]
    enable_notification_channel = StackdriverEnableNotificationChannelsOperator(
        task_id="enable-notification-channel", filter_='type="pubsub"'
    )
    # [END howto_operator_gcp_stackdriver_enable_notification_channel]

    # [START howto_operator_gcp_stackdriver_disable_notification_channel]
    disable_notification_channel = StackdriverDisableNotificationChannelsOperator(
        task_id="disable-notification-channel", filter_=f'displayName="{CHANNEL_1_NAME}"'
    )
    # [END howto_operator_gcp_stackdriver_disable_notification_channel]

    # [START howto_operator_gcp_stackdriver_list_notification_channel]
    list_notification_channel = StackdriverListNotificationChannelsOperator(
        task_id="list-notification-channel", filter_='type="pubsub"'
    )
    # [END howto_operator_gcp_stackdriver_list_notification_channel]

    # [START howto_operator_gcp_stackdriver_upsert_alert_policy]
    create_alert_policy = StackdriverUpsertAlertOperator(
        task_id="create-alert-policies",
        alerts=json.dumps({"policies": [TEST_ALERT_POLICY_1, TEST_ALERT_POLICY_2]}),
    )
    # [END howto_operator_gcp_stackdriver_upsert_alert_policy]

    # [START howto_operator_gcp_stackdriver_enable_alert_policy]
    enable_alert_policy = StackdriverEnableAlertPoliciesOperator(
        task_id="enable-alert-policies",
        filter_=f'(displayName="{ALERT_1_NAME}" OR displayName="{ALERT_2_NAME}")',
    )
    # [END howto_operator_gcp_stackdriver_enable_alert_policy]

    # [START howto_operator_gcp_stackdriver_disable_alert_policy]
    disable_alert_policy = StackdriverDisableAlertPoliciesOperator(
        task_id="disable-alert-policies",
        filter_=f'displayName="{ALERT_1_NAME}"',
    )
    # [END howto_operator_gcp_stackdriver_disable_alert_policy]

    # [START howto_operator_gcp_stackdriver_list_alert_policy]
    list_alert_policies = StackdriverListAlertPoliciesOperator(
        task_id="list-alert-policies",
    )
    # [END howto_operator_gcp_stackdriver_list_alert_policy]

    # [START howto_operator_gcp_stackdriver_delete_notification_channel]
    delete_notification_channel = StackdriverDeleteNotificationChannelOperator(
        task_id="delete-notification-channel",
        name="{{ task_instance.xcom_pull('list-notification-channel')[0]['name'] }}",
    )
    # [END howto_operator_gcp_stackdriver_delete_notification_channel]
    delete_notification_channel.trigger_rule = TriggerRule.ALL_DONE

    delete_notification_channel_2 = StackdriverDeleteNotificationChannelOperator(
        task_id="delete-notification-channel-2",
        name="{{ task_instance.xcom_pull('list-notification-channel')[1]['name'] }}",
        trigger_rule=TriggerRule.ALL_DONE,
    )

    # [START howto_operator_gcp_stackdriver_delete_alert_policy]
    delete_alert_policy = StackdriverDeleteAlertOperator(
        task_id="delete-alert-policy",
        name="{{ task_instance.xcom_pull('list-alert-policies')[0]['name'] }}",
    )
    # [END howto_operator_gcp_stackdriver_delete_alert_policy]
    delete_alert_policy.trigger_rule = TriggerRule.ALL_DONE

    delete_alert_policy_2 = StackdriverDeleteAlertOperator(
        task_id="delete-alert-policy-2",
        name="{{ task_instance.xcom_pull('list-alert-policies')[1]['name'] }}",
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        create_notification_channel,
        enable_notification_channel,
        disable_notification_channel,
        list_notification_channel,
        create_alert_policy,
        enable_alert_policy,
        disable_alert_policy,
        list_alert_policies,
        delete_notification_channel,
        delete_notification_channel_2,
        delete_alert_policy,
        delete_alert_policy_2,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import asyncio
import logging
from unittest import mock

import pytest
from google.cloud.bigquery_datatransfer_v1 import TransferState

from airflow.providers.google.cloud.triggers.bigquery_dts import BigQueryDataTransferRunTrigger
from airflow.triggers.base import TriggerEvent

PROJECT_ID = "test-project-id"
CONFIG_ID = "test-config-id"
RUN_ID = "test-run-id"
POLL_INTERVAL = 10
GCP_CONN_ID = "google-cloud-default-id"
LOCATION = "us-central1"
IMPERSONATION_CHAIN = ["test", "chain"]


@pytest.fixture
def trigger():
    trigger = BigQueryDataTransferRunTrigger(
        project_id=PROJECT_ID,
        config_id=CONFIG_ID,
        run_id=RUN_ID,
        poll_interval=POLL_INTERVAL,
        gcp_conn_id=GCP_CONN_ID,
        location=LOCATION,
        impersonation_chain=IMPERSONATION_CHAIN,
    )
    return trigger


class TestBigQueryDataTransferRunTrigger:
    TRIGGER_MODULE_PATH = "airflow.providers.google.cloud.triggers.bigquery_dts"
    TRIGGER_PATH = f"{TRIGGER_MODULE_PATH}.BigQueryDataTransferRunTrigger"

    def test_serialize(self, trigger):
        classpath, kwargs = trigger.serialize()
        assert classpath == self.TRIGGER_PATH
        assert kwargs == {
            "project_id": PROJECT_ID,
            "config_id": CONFIG_ID,
            "run_id": RUN_ID,
            "poll_interval": POLL_INTERVAL,
            "gcp_conn_id": GCP_CONN_ID,
            "location": LOCATION,
            "impersonation_chain": IMPERSONATION_CHAIN,
        }

    @pytest.mark.parametrize(
        "attr, expected_value",
        [
            ("gcp_conn_id", GCP_CONN_ID),
            ("location", LOCATION),
            ("impersonation_chain", IMPERSONATION_CHAIN),
        ],
    )
    def test_get_async_hook(self, attr, expected_value, trigger):
        hook = trigger._get_async_hook()
        actual_value = hook._hook_kwargs.get(attr)
        assert actual_value is not None
        assert actual_value == expected_value

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_MODULE_PATH}.AsyncBiqQueryDataTransferServiceHook.get_transfer_run")
    async def test_run_returns_success_event(self, mock_hook, trigger):
        mock_hook.return_value = mock.MagicMock(state=TransferState.SUCCEEDED)
        expected_event = TriggerEvent(
            {
                "run_id": RUN_ID,
                "status": "success",
                "message": "Job completed",
                "config_id": CONFIG_ID,
            }
        )
        actual_event = await trigger.run().asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_MODULE_PATH}.AsyncBiqQueryDataTransferServiceHook.get_transfer_run")
    async def test_run_returns_failed_event(self, mock_hook, trigger):
        mock_hook.return_value = mock.MagicMock(state=TransferState.FAILED)
        expected_event = TriggerEvent(
            {
                "status": "failed",
                "run_id": RUN_ID,
                "message": "Job has failed",
            }
        )
        actual_event = await trigger.run().asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_MODULE_PATH}.AsyncBiqQueryDataTransferServiceHook.get_transfer_run")
    async def test_run_returns_exception_event(self, mock_hook, trigger):
        error_msg = "test error msg"
        mock_hook.side_effect = Exception(error_msg)
        expected_event = TriggerEvent(
            {
                "status": "failed",
                "message": f"Trigger failed with exception: {error_msg}",
            }
        )
        actual_event = await trigger.run().asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_MODULE_PATH}.AsyncBiqQueryDataTransferServiceHook.get_transfer_run")
    async def test_run_returns_cancelled_event(self, mock_hook, trigger):
        mock_hook.return_value = mock.MagicMock(state=TransferState.CANCELLED)
        expected_event = TriggerEvent(
            {
                "status": "cancelled",
                "run_id": RUN_ID,
                "message": "Job was cancelled",
            }
        )
        actual_event = await trigger.run().asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_MODULE_PATH}.AsyncBiqQueryDataTransferServiceHook.get_transfer_run")
    async def test_run_loop_is_still_running(self, mock_hook, trigger, caplog):
        mock_hook.return_value = mock.MagicMock(state=TransferState.RUNNING)

        caplog.set_level(logging.INFO)

        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        assert not task.done()
        assert f"Current job status is: {TransferState.RUNNING}"
        assert f"Sleeping for {POLL_INTERVAL} seconds."



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG for testing Google Dataflow Beam Pipeline Operator with Java.

Important Note:
    This test downloads Java JAR file from the public bucket. In case the JAR file cannot be downloaded
    or is not compatible with the Java version used in the test, the source code for this test can be
    downloaded from here (https://beam.apache.org/get-started/wordcount-example) and needs to be compiled
    manually in order to work.

    You can follow the instructions on how to pack a self-executing jar here:
    https://beam.apache.org/documentation/runners/dataflow/

Requirements:
    These operators require the gcloud command and Java's JRE to run.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.apache.beam.hooks.beam import BeamRunnerType
from airflow.providers.apache.beam.operators.beam import BeamRunJavaPipelineOperator
from airflow.providers.google.cloud.operators.dataflow import CheckJobRunning
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.transfers.gcs_to_local import GCSToLocalFilesystemOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
DAG_ID = "dataflow_native_java"

BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}"
PUBLIC_BUCKET = "airflow-system-tests-resources"

JAR_FILE_NAME = "word-count-beam-bundled-0.1.jar"
# For the distributed system, we need to store the JAR file in a folder that can be accessed by multiple
# worker.
# For example in Composer the correct path is gcs/data/word-count-beam-bundled-0.1.jar.
# Because gcs/data/ is shared folder for Airflow's workers.
IS_COMPOSER = bool(os.environ.get("COMPOSER_ENVIRONMENT", ""))
LOCAL_JAR = f"gcs/data/{JAR_FILE_NAME}" if IS_COMPOSER else JAR_FILE_NAME
REMOTE_JAR_FILE_PATH = f"dataflow/java/{JAR_FILE_NAME}"
GCS_JAR = f"gs://{PUBLIC_BUCKET}/dataflow/java/{JAR_FILE_NAME}"
GCS_OUTPUT = f"gs://{BUCKET_NAME}"
LOCATION = "europe-west3"

with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "dataflow", "java"],
) as dag:
    create_bucket = GCSCreateBucketOperator(task_id="create_bucket", bucket_name=BUCKET_NAME)

    download_file = GCSToLocalFilesystemOperator(
        task_id="download_file",
        object_name=REMOTE_JAR_FILE_PATH,
        bucket=PUBLIC_BUCKET,
        filename=LOCAL_JAR,
    )

    # [START howto_operator_start_java_job_local_jar]
    start_java_job_direct = BeamRunJavaPipelineOperator(
        task_id="start_java_job_direct",
        jar=LOCAL_JAR,
        pipeline_options={
            "output": GCS_OUTPUT,
        },
        job_class="org.apache.beam.examples.WordCount",
        dataflow_config={
            "check_if_running": CheckJobRunning.WaitForRun,
            "location": LOCATION,
            "poll_sleep": 10,
        },
    )
    # [END howto_operator_start_java_job_local_jar]

    start_java_job_direct_deferrable = BeamRunJavaPipelineOperator(
        task_id="start_java_job_direct_deferrable",
        jar=GCS_JAR,
        pipeline_options={
            "output": GCS_OUTPUT,
        },
        job_class="org.apache.beam.examples.WordCount",
        dataflow_config={
            "check_if_running": CheckJobRunning.WaitForRun,
            "location": LOCATION,
            "poll_sleep": 10,
        },
        deferrable=True,
    )

    # [START howto_operator_start_java_job_jar_on_gcs]
    start_java_job_dataflow = BeamRunJavaPipelineOperator(
        runner=BeamRunnerType.DataflowRunner,
        task_id="start_java_job_dataflow",
        jar=GCS_JAR,
        pipeline_options={
            "output": GCS_OUTPUT,
        },
        job_class="org.apache.beam.examples.WordCount",
        dataflow_config={
            "job_name": f"java-pipeline-job-{ENV_ID}",
            "check_if_running": CheckJobRunning.IgnoreJob,
            "location": LOCATION,
            "poll_sleep": 10,
            "append_job_name": False,
        },
    )
    # [END howto_operator_start_java_job_jar_on_gcs]

    # [START howto_operator_start_java_job_jar_on_gcs_deferrable]
    start_java_job_dataflow_deferrable = BeamRunJavaPipelineOperator(
        runner=BeamRunnerType.DataflowRunner,
        task_id="start_java_job_dataflow_deferrable",
        jar=GCS_JAR,
        pipeline_options={
            "output": GCS_OUTPUT,
        },
        job_class="org.apache.beam.examples.WordCount",
        dataflow_config={
            "job_name": f"deferrable-java-pipeline-job-{ENV_ID}",
            "check_if_running": CheckJobRunning.WaitForRun,
            "location": LOCATION,
            "poll_sleep": 10,
            "append_job_name": False,
        },
        deferrable=True,
    )
    # [END howto_operator_start_java_job_jar_on_gcs_deferrable]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME, trigger_rule=TriggerRule.ALL_DONE
    )

    (
        # TEST SETUP
        create_bucket
        >> download_file
        # TEST BODY
        >> [
            start_java_job_direct,
            start_java_job_direct_deferrable,
            start_java_job_dataflow,
            start_java_job_dataflow_deferrable,
        ]
        # TEST TEARDOWN
        >> delete_bucket
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)
